package ppi.prediction;

import cern.colt.matrix.linalg.Algebra;
import cern.colt.matrix.DoubleMatrix2D;
import cern.colt.matrix.impl.DenseDoubleMatrix2D;

import java.util.Random;

public class VariationalLogistic {

    private static int MAX_ITER = 100;
    private static int MAX_ITER_PRED = 20;

    double data[][];
    int    classes[], numVars;

    double paramPriorMean[], invParamPriorCov[][], invSigmaPriorMuPrior[];
    double paramPosteriorCov[][], invParamPosteriorCov[][], paramPosteriorSecondMoments[][];
    double paramPosteriorMean[];
    double paramPosteriorMeanAux[];

    double score;

    double scratchPad1[][], scratchPad2[][];

    public VariationalLogistic(double data[][], int classes[], double paramPriorMean[], double invParamPriorCov[][]) {
        this.score = -Double.MAX_VALUE;
        this.numVars = paramPriorMean.length;
        this.invParamPriorCov = invParamPriorCov;
        this.paramPriorMean = paramPriorMean;
        this.invSigmaPriorMuPrior = new double[this.numVars];
        this.paramPosteriorMeanAux = new double[this.numVars];
        if (data != null) {
            this.data = data;
            this.classes = classes;
            this.paramPosteriorMean = new double[this.numVars];
            this.paramPosteriorSecondMoments = new double[this.numVars][this.numVars];
            this.invParamPosteriorCov = new double[this.numVars][this.numVars];
            this.invSigmaPriorMuPrior = new double[this.numVars];
        } else {
            this.paramPosteriorMean = paramPriorMean;
            this.invParamPosteriorCov = invParamPriorCov;
        }
        this.scratchPad1 = new double[this.numVars][this.numVars];
        this.scratchPad2 = new double[this.numVars][this.numVars];
    }

    public void updatePosterior() {
        Random r = new Random();
        double qsi, lambdaQsi;
        double workingInvCovPrior[][] = new double[this.numVars][this.numVars];
        double workingMeanPrior[] = new double[this.numVars];
        System.arraycopy(this.paramPriorMean, 0, workingMeanPrior, 0, this.numVars);
        for (int i = 0; i < this.numVars; i++)
            System.arraycopy(this.invParamPriorCov[i], 0, workingInvCovPrior[i], 0, this.numVars);
        for (int d = 0; d < this.data.length; d++) {
            System.out.println("Processing data point " + (d + 1));
            qsi = r.nextDouble() * 10.;
            lambdaQsi = Math.tanh(qsi * 0.5) / (4. * qsi);
            //System.out.println(" -- qsi 0 = " + qsi);
            for (int v1 = 0; v1 < this.numVars; v1++) {
                this.paramPosteriorMean[v1] = r.nextGaussian();
                for (int v2 = 0; v2 < this.numVars; v2++)
                    this.invParamPosteriorCov[v1][v2] = 0.;
                this.invParamPosteriorCov[v1][v1] = r.nextDouble() + 0.5;
            }
            this.paramPosteriorCov = inverse(this.invParamPosteriorCov);
            for (int v1 = 0; v1 < this.numVars; v1++)
                for (int v2 = 0; v2 < this.numVars; v2++)
                    this.paramPosteriorSecondMoments[v1][v2] = this.paramPosteriorCov[v1][v2] + this.paramPosteriorMean[v1] * this.paramPosteriorMean[v2];
            //System.out.println(" - logTrainingBound 0 = " + logTrainingBound(d, qsi, lambdaQsi, workingMeanPrior, workingInvCovPrior));
            for (int i = 0; i < VariationalLogistic.MAX_ITER; i++) {
                //System.out.print("Iteration " + (i + 1) + "... ");
                for (int v1 = 0; v1 < this.numVars; v1++)
                    for (int v2 = 0; v2 < this.numVars; v2++)
                        this.invParamPosteriorCov[v1][v2] = workingInvCovPrior[v1][v2] + 2. * lambdaQsi *
                                                            this.data[d][v1] * this.data[d][v2];
                this.paramPosteriorCov = inverse(this.invParamPosteriorCov);
                for (int j = 0; j < this.numVars; j++) {
                    this.invSigmaPriorMuPrior[j] = 0.;
                    for (int k = 0; k < this.numVars; k++)
                        this.invSigmaPriorMuPrior[j] += workingInvCovPrior[j][k] * workingMeanPrior[k];
                }
                for (int v = 0; v < this.numVars; v++)
                    this.paramPosteriorMeanAux[v] = this.invSigmaPriorMuPrior[v] +
                                                    (this.classes[d] - 0.5) * this.data[d][v];
                for (int v1 = 0; v1 < this.numVars; v1++) {
                    this.paramPosteriorMean[v1] = 0.;
                    for (int v2 = 0; v2 < this.numVars; v2++)
                        this.paramPosteriorMean[v1] += this.paramPosteriorCov[v1][v2] * this.paramPosteriorMeanAux[v2];
                }
                for (int v1 = 0; v1 < this.numVars; v1++)
                    for (int v2 = 0; v2 < this.numVars; v2++)
                        this.paramPosteriorSecondMoments[v1][v2] = this.paramPosteriorCov[v1][v2] +
                                                                   this.paramPosteriorMean[v1] * this.paramPosteriorMean[v2];
                qsi = 0.;
                for (int v1 = 0; v1 < this.numVars; v1++)
                    for (int v2 = 0; v2 < this.numVars; v2++)
                        qsi += this.paramPosteriorSecondMoments[v1][v2] * this.data[d][v1] * this.data[d][v2];
                qsi = Math.sqrt(qsi);
                lambdaQsi = Math.tanh(qsi * 0.5) / (4. * qsi);
                //System.out.println(" -- qsi = " + qsi);
                //System.out.println(" - logTrainingBound = " + logTrainingBound(d, qsi, lambdaQsi, workingMeanPrior, workingInvCovPrior));
            }
            System.arraycopy(this.paramPosteriorMean, 0, workingMeanPrior, 0, this.numVars);
            for (int j = 0; j < this.numVars; j++)
                System.arraycopy(this.invParamPosteriorCov[j], 0, workingInvCovPrior[j], 0, this.numVars);
        }
        //System.exit(0);
    }

    public double logPredictiveBound(double newPoint[], int newClass) {
        Random r = new Random();
        double qsi, lambdaQsi;
        double predictiveInvCov[][] = new double[this.numVars][this.numVars];
        double predictiveMean[] = new double[this.numVars];
        double predictiveSecondMoments[][] = new double[this.numVars][this.numVars];

        qsi = r.nextDouble() * 10.;
        lambdaQsi = Math.tanh(qsi * 0.5) / (4. * qsi);
        for (int i = 0; i < VariationalLogistic.MAX_ITER_PRED; i++) {
            for (int v1 = 0; v1 < this.numVars; v1++)
                for (int v2 = 0; v2 < this.numVars; v2++)
                    predictiveInvCov[v1][v2] = this.invParamPosteriorCov[v1][v2] + 2. * lambdaQsi *
                                               newPoint[v1] * newPoint[v2];
            double predictiveCov[][] = inverse(predictiveInvCov);
            for (int j = 0; j < this.numVars; j++) {
                this.invSigmaPriorMuPrior[j] = 0.;
                for (int k = 0; k < this.numVars; k++)
                    this.invSigmaPriorMuPrior[j] += this.invParamPosteriorCov[j][k] * this.paramPosteriorMean[k];
            }
            for (int v = 0; v < this.numVars; v++)
                this.paramPosteriorMeanAux[v] = this.invSigmaPriorMuPrior[v] + (newClass - 0.5) * newPoint[v];
            for (int v1 = 0; v1 < this.numVars; v1++) {
                predictiveMean[v1] = 0.;
                for (int v2 = 0; v2 < this.numVars; v2++)
                    predictiveMean[v1] += predictiveCov[v1][v2] * this.paramPosteriorMeanAux[v2];
            }
            for (int v1 = 0; v1 < this.numVars; v1++)
                for (int v2 = 0; v2 < this.numVars; v2++)
                    predictiveSecondMoments[v1][v2] = predictiveCov[v1][v2] + predictiveMean[v1] * predictiveMean[v2];
            qsi = 0.;
            for (int v1 = 0; v1 < this.numVars; v1++)
                for (int v2 = 0; v2 < this.numVars; v2++)
                    qsi += predictiveSecondMoments[v1][v2] * newPoint[v1] * newPoint[v2];
            qsi = Math.sqrt(qsi);
            lambdaQsi = Math.tanh(qsi * 0.5) / (4. * qsi);
        }

        double output = -Math.log(1. + Math.exp(-qsi)) - 0.5 * qsi + lambdaQsi * qsi * qsi;
        output += 0.5 * (Math.log(determinant(this.invParamPosteriorCov)) - Math.log(determinant(predictiveInvCov)));

        //NOTE: DEPENDING ON THE PRIOR, THE LINE ABOVE MIGHT BE NUMERICALLY UNSTABLE. USE THE CODE BELOW INSTEAD.
        //IF THE PRIOR IS HEAVILY CONCENTRATED AROUND THE MEAN, SCALE MIGHT NEED TO BE SET TO A SMALL NUMBER (E.G.,
        //SCALE = 1 / 100.)
        //
        //double scale = 100;
        //for (int i = 0; i < this.numVars; i++)
        //    for (int j = 0; j < this.numVars; j++) {
        //        this.scratchPad1[i][j] = this.invParamPosteriorCov[i][j] * scale;
        //        this.scratchPad2[i][j] = predictiveInvCov[i][j] * scale;
        //    }
        //output += 0.5 * (Math.log(determinant(this.scratchPad1)) - Math.log(determinant(this.scratchPad2)));

        if (Double.isNaN(output)) {
            System.out.println("Output NaN");
            System.exit(0);
        }

        if (Double.isInfinite(output)) {
            System.out.println("Output Infinite");
            System.exit(0);
        }


        double productAux[] = new double[this.numVars];
        for (int i = 0; i < this.numVars; i++) {
            productAux[i] = 0;
            for (int j = 0; j < this.numVars; j++)
                productAux[i] += predictiveMean[j] * predictiveInvCov[j][i];
        }
        for (int i = 0; i < this.numVars; i++)
            output += 0.5 * productAux[i] * predictiveMean[i];
        for (int i = 0; i < this.numVars; i++) {
            productAux[i] = 0;
            for (int j = 0; j < this.numVars; j++)
                productAux[i] += this.paramPosteriorMean[j] * this.invParamPosteriorCov[j][i];
        }
        for (int i = 0; i < this.numVars; i++) {
            output -= 0.5 * productAux[i] * this.paramPosteriorMean[i];
            if (Double.isNaN(output)) {
                System.out.println("Output NaN at " + i);
                System.exit(0);
            }
        }
        return output;
    }

    /**
     * Extra auxiliary methods
     */

    public int[] prediction(int test[][]) {
        int output[] = new int[test.length];
        for (int i = 0; i < test.length; i++) {
            double thetaX = 0.;
            for (int j = 0; j < test[i].length; j++)
                thetaX += this.paramPosteriorMean[test[i][j]];
            if (thetaX > 0.)
                output[i] = 1;
            else
                output[i] = 0;
        }
        return output;
    }

    public double[][] getPosteriorCov() {
        return this.paramPosteriorCov;
    }

    public double[][] getInvPosteriorCov() {
        return this.invParamPosteriorCov;
    }

    public double[] getPosteriorMean() {
        return this.paramPosteriorMean;
    }

    protected static double[][] inverse(double matrix[][]) {
        Algebra algebra = new Algebra();
        DoubleMatrix2D m = new DenseDoubleMatrix2D(matrix);
        DoubleMatrix2D result = algebra.inverse(m);
        return result.toArray();
    }

    public static double determinant(double[][] m) {
        return new Algebra().det(new DenseDoubleMatrix2D(m));
    }

    /**
     * Calculate MLE by Newton's method and then dump it on a text file. It assumes an intercept term is
     * already included in the data. It also allows for a different weight for the negative (zero) class.
     * Classes should be enconded as {0, 1} or {-1, 1}.
     */

    public static double[] mle(double X[][], double y[], double negWeight) {
        int numVars = X[0].length;
        int sampleSize = X.length;
        int maxIter = 15;

        for (int d = 0; d < y.length; d++)
            if (y[d] < 0.)
                y[d] = 0.;

        double pointWeight[] = new double[sampleSize];
        for (int d = 0; d < sampleSize; d++) {
            if (y[d] == 1)
                pointWeight[d] = 1.;
            else
                pointWeight[d] = negWeight;
        }

        double z[] = new double[sampleSize];
        double p[] = new double[sampleSize];
        double W[] = new double[sampleSize];
        double beta[] = new double[numVars];
        double XtW[][] = new double[numVars][sampleSize];
        double XtWX[][] = new double[numVars][numVars];
        double innerProd[] = new double[sampleSize];

        for (int v = 0; v < numVars; v++)
            beta[v] = 0.;
        for (int d = 0; d < sampleSize; d++) {
            innerProd[d] = 0.;
            for (int v = 0; v < numVars; v++)
                innerProd[d] += beta[v] * X[d][v];
        }

        System.out.println("Starting MLE...");
        double logLike = 0.;
        for (int d = 0; d < sampleSize; d++) {
            if (y[d] > 0)
               logLike -= pointWeight[d] * Math.log(1 + Math.exp(-innerProd[d]));
            else
               logLike -= pointWeight[d] * Math.log(1 + Math.exp(innerProd[d]));
            if (Double.isNaN(logLike))
                System.out.println("NaN here");
        }
        System.out.println("Log-likelihood (0) = " + logLike);

        double betaBackup[] = new double[beta.length];
        mainBetaIter:
        for (int i = 0; i < maxIter; i++) {
            System.arraycopy(beta, 0, betaBackup, 0, beta.length);
            for (int d = 0; d < sampleSize; d++) {
                p[d] = sigmoid(innerProd[d]);
                W[d] = p[d] * (1. - p[d]);
                z[d] = (y[d] - p[d]) / W[d];
                for (int v = 0; v < numVars; v++)
                    z[d] += X[d][v] * beta[v];
            }

            double altXtWX[][] = new double[numVars][numVars];
            for (int v1 = 0; v1 < numVars; v1++)
                for (int v2 = 0; v2 < numVars; v2++)
                    altXtWX[v1][v2] = 0.;
            System.out.println("Stage 1");
            for (int d = 0; d < sampleSize; d++)
                for (int v1 = 0; v1 < numVars; v1++)
                    for (int v2 = v1; v2 < numVars; v2++)
                        altXtWX[v1][v2] += X[d][v1] * X[d][v2] * W[d];
            for (int v1 = 0; v1 < numVars; v1++)
                for (int v2 = v1 + 1; v2 < numVars; v2++)
                    altXtWX[v2][v1] = altXtWX[v1][v2];
            System.out.println("Stage 2");
            for (int v = 0; v < numVars; v++)
                for (int d = 0; d < sampleSize; d++)
                    XtW[v][d] = X[d][v] * W[d];
            System.out.println("Stage 3");
            for (int v1 = 0; v1 < numVars; v1++)
                for (int v2 = v1; v2 < numVars; v2++)
                     XtWX[v1][v2] = 0.;
            for (int d = 0; d < sampleSize; d++)
                for (int v1 = 0; v1 < numVars; v1++)
                    for (int v2 = v1; v2 < numVars; v2++)
                        XtWX[v1][v2] += XtW[v1][d] * X[d][v2];
            for (int v1 = 0; v1 < numVars; v1++)
                for (int v2 = v1 + 1; v2 < numVars; v2++)
                    XtWX[v2][v1] = XtWX[v1][v2];

            double invXtWX[][] = inverse(XtWX);
            System.out.println("Stage 4");
            for (int v = 0; v < numVars; v++) {
                beta[v] = 0.;
                for (int v0 = 0; v0 < numVars; v0++)
                    for (int d = 0; d < sampleSize; d++)
                         beta[v] += invXtWX[v0][v] * XtW[v0][d] * z[d];
            }

            System.out.println("Stage 5");
            for (int d = 0; d < sampleSize; d++) {
                innerProd[d] = 0.;
                for (int v = 0; v < numVars; v++)
                    innerProd[d] += beta[v] * X[d][v];
            }
            System.out.println("Stage 6");
            logLike = 0.;
            for (int d = 0; d < sampleSize; d++) {
                if (y[d] > 0)
                   logLike -= pointWeight[d] * Math.log(1 + Math.exp(-innerProd[d]));
                else
                   logLike -= pointWeight[d] * Math.log(1 + Math.exp(innerProd[d]));
                if (Double.isInfinite(logLike)) {
                    System.out.println("! Infinite");
                    beta = betaBackup;
                    break mainBetaIter;
                }
                if (Double.isNaN(logLike)) {
                    System.out.println("! NaN");
                    beta = betaBackup;
                    break mainBetaIter;
                }
            }
            System.out.println("Log-likelihood (" + (i + 1) + ") = " + logLike);
        }
        return beta;
    }

    protected static double sigmoid(double x) {
        return 1. / (1. + Math.exp(-x));
    }

    public double getScore() {
        return this.score;
    }
}